import os
import copy
import shutil
from solo.logger import Logger
from solo.template import TemplateConstants, Template
from solo.basic_files import user_home_directory
from solo.factory import create_factory
from solo.configuration.configuration import Configuration
from .stealth import Stealth


logger = Logger(__name__)


class DBManager(Configuration):

    """
        DBManager has two main roles:
            - be aware of which databases are available in a particular system
              using the configuration file
            - distribute queries to the active database(s)
    """

    db_factory = create_factory('DB')
    journal = []

    def __init__(self, parent):
        super().__init__(self.read_config(), environment=True)
        self._parent = parent
        for name, data in self.databases.items():
            self[name] = self.db_factory.create(data['class'], name, data)
        # since we use the base loader in yaml, booleans are not
        # interpreted, just a quick walk through them.
        for name, database in self.databases.items():
            if 'cache_fetch' in database:
                cache = database['cache_fetch'].lower()
                database['cache_fetch'] = (cache == 'yes') or (cache == 'true')

    @staticmethod
    def read_config():
        if os.environ.get('R2D2_CONFIG') is not None:
            config_file = os.environ['R2D2_CONFIG']
        else:
            path = os.path.join(user_home_directory(), '.r2d2')
            config_file = os.path.join(path, 'config.yaml')
            if not os.path.exists(config_file):
                os.makedirs(path, exist_ok=True)
                source = os.path.join(os.path.dirname(__file__), '../config.yaml')
                shutil.copy(source, config_file)
        return config_file

    def fetch_journal(self):
        return self._fetch_journal

    def store_journal(self):
        return self._store_journal

    def cache(self):
        return self.get('cache_name', None)

    def fetch_data(self, source, target, request, archive_side_request):
        missing = False
        cache_fetch = False
        if 'database' in request and request['database'] != 'default':
            databases = [request['database']]
        else:
            databases = [x for x in self.fetch_order]
        done = False
        i = 0
        while not done and i < len(databases):
            db_name = databases[i]
            database = self[db_name]
            cache_fetch = self.databases[db_name].get('cache_fetch', False)
            try:
                source, target = database.fetch_data(source, target, request, archive_side_request)
            except FileNotFoundError as e:
                pass
            else:
                done = True
            i += 1
        if done:
            if cache_fetch:
                self.cache_data(target, request, archive_side_request)
        else:
            missing = True
            if not request.ignore_missing:
                raise FileNotFoundError(f'Could not find {source} in {" or ".join(self.fetch_order)}')
            else:
                logger.warning(f'Could not find {source}, ignore_missing is set to "yes"')
        target = os.path.abspath(target)
        self.journal.append(('fetch', source, target, request, cache_fetch, missing))

    def store_data(self, source, target, request, archive_side_request):
        missing = False
        store_path = target
        try:
            if 'database' in request and request['database'] != 'default':
                _, store_path = self[request['database']].store_data(source, target, request, archive_side_request)
            else:
                for db_name in self.store_order:
                    _, store_path = self[db_name].store_data(source, target, request, archive_side_request)
        except FileNotFoundError:
            missing = True
            if not request.ignore_missing:
                raise
            else:
                logger.warning(f'Could not find {source}, ignore_missing is set to "yes"')
        source = os.path.abspath(source)
        self.journal.append(('store', source, store_path, request, False, missing))

    def cache_data(self, source, request, archive_side_request):
        """
            Have to do some "dirty work" here, but this is possibly the
            simplest way to handle the cache.
        """
        template = self._parent.get_template(request)
        target = Template.substitute_string(template, TemplateConstants.DOLLAR_PARENTHESES, request.get)

        db = Stealth.silence(self[self.cache()])
        db.store_data(source, target, request, archive_side_request)

    @classmethod
    def reset_journal(cls):
        cls.journal = []


